#include <stdio.h>
#include <stdlib.h>

//Intel
#define PREFETCH " prefetchnta "
#define FEMMS ""
#define EMMS ""
#define SFENCE " sfence\n"

//AMD
//#define PREFETCH " prefetch "
//#define FEMMS " femms\n"
//#define EMMS " emms\n"
//#define SFENCE " sfence\n"

#define NAME(a) \
unsigned int a##csum(const unsigned char * buff, int len, \
			unsigned int sum); \
unsigned int a##copy(const char *src, char *dst, \
                        int len, int sum, int *src_err_ptr, int *dst_err_ptr)

/* This makes adding/removing test functions easier
 */
/* asm ones... */
NAME(kernel_);
NAME(kernel2_);
NAME(kernelpii_);
NAME(kernelpiipf_);
NAME(kpf_);
/* and C */
#include "csum_pfm.c"
#include "csum_pfm2.c"
#include "copy_ntq.c"
#include "copy_ntqpf.c"
#include "copy_ntqpf2.c"
#include "copy_ntqpfm.c"

#define TRY_TIMES 32
#define BUFSZ   (4 * 1024*1024)
#define CHUNKSZ (1 * 1024)
#define POISON	1	// want to check correctness?
#define MAX_START_OFS	32
#define MAX_END_OFS	32

#define CHUNKCNT (BUFSZ/2 / CHUNKSZ)
#define BUF_ALIGN 0x10

char *buffer;
char *chunk_addr[CHUNKCNT];
int start_ofs[CHUNKCNT];
int end_ofs[CHUNKCNT];

typedef unsigned int csum_func(const unsigned char * buff, int len,
		unsigned int sum);
typedef unsigned int copy_func(const char *src, char *dst,
		int len, int sum, int *src_err_ptr, int *dst_err_ptr);

static inline long long rdtsc()
{
	unsigned int low,high;
	__asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
;
	return low + (((long long)high)<<32);
}

int die(const char *msg) {
	puts(msg);
	abort();
	return 1;
}

void prepare() {
	int i;

	srand(time(0));

	//ramdom buffer contents
        for(i=0; i<BUFSZ; i++) {
		buffer[i] = rand();
	}
	//linear chunk addresses...
        for(i=0; i<CHUNKCNT; i++) {
		chunk_addr[i] = buffer + CHUNKSZ*i;
	}
	//randomize 'em
        for(i=0; i<CHUNKCNT-1; i++) {
		int r = i+rand()%(CHUNKCNT-i);
		char *tp = chunk_addr[i];
		chunk_addr[i] = chunk_addr[r];
		chunk_addr[r] = tp;
	}
	//ramdom chunk offsets
        for(i=0; i<CHUNKCNT; i++) {
		// start must be even (really?)
		start_ofs[i] = (rand() % (MAX_START_OFS/2+1))<<1;
		end_ofs[i] = rand() % (MAX_END_OFS+1);
		//start_ofs[i] = 14;
		//end_ofs[i] = 16;
	}
}

unsigned test_one_csum(csum_func *func, char *name) {
	int i,j;
	unsigned long long before,after,min,max;
	unsigned sum,sum2;

	printf("%20s - ", name);
	
	// dummy run to ensure coherent mem/cache state
	sum = 0;
	for(j=0; j<CHUNKCNT; j++) {
		char *addr = chunk_addr[j]+start_ofs[j];
		int sz = CHUNKSZ-start_ofs[j]-end_ofs[j];
		sum = func(addr,sz,sum);
		//sum = func(chunk_addr[j], CHUNKSZ, sum);
//debug printf("addr:0x%08x sz:%i\n",addr,sz);
	}

	// pick fastest run
	min = ~0ULL;
	max = 0;
	for (i=0;i<TRY_TIMES;i++) {
		sum2 = 0;
		before = rdtsc();
    		for(j=0; j<CHUNKCNT; j++) {
			char *addr = chunk_addr[j]+start_ofs[j];
			int sz = CHUNKSZ-start_ofs[j]-end_ofs[j];
			sum2 = func(addr,sz,sum2);
			//sum2 = func(chunk_addr[j], CHUNKSZ, sum2);
		}
		after = rdtsc();
		if(sum != sum2) die("bad sum");
		if(before > after) die("timer overflow");
		after -= before;
		if(min > after) min = after;
		if(max < after) max = after;
	}
	printf("took %5lli max,%5lli min cycles per kb. sum=0x%08x\n",
		max / (BUFSZ/2 / 1024),
		min / (BUFSZ/2 / 1024),
		sum
	);
	return sum;
}

unsigned test_one_copy(copy_func *func, char *name) {
	int i,j;
	unsigned long long before,after,min,max;
	unsigned sum,sum2;
	int err;

	printf("%20s - ", name);
	
	// dummy run to ensure coherent mem/cache state
	if(POISON) memset(buffer+BUFSZ/2,0xaa,BUFSZ/2);
	sum = 0;
	for(j=0; j<CHUNKCNT; j++) {
		char *addr = chunk_addr[j]+start_ofs[j];
		int sz = CHUNKSZ-start_ofs[j]-end_ofs[j];
		sum = func(addr,addr+BUFSZ/2,sz,sum,&err,&err);
		//sum = func(chunk_addr[j],chunk_addr[j]+BUFSZ/2,
		//	CHUNKSZ,sum,&err,&err);
		if(POISON) {
			if(memcmp(addr,addr+BUFSZ/2,sz)!=0) die("bad copy");
			if(start_ofs[j] && addr[BUFSZ/2-1]!=(char)0xaa) {
				die("error at start");
			}
			if(end_ofs[j] && addr[BUFSZ/2+sz]!=(char)0xaa)
				die("error at end");
		}

	}
	//if(POISON) if(memcmp(buffer,buffer+BUFSZ/2,BUFSZ/2)!=0) die("bad copy");

	// pick fastest run
	min = ~0ULL;
	max = 0;
	for (i=0; i<TRY_TIMES; i++) {
		sum2 = 0;
		before = rdtsc();
		for(j=0; j<CHUNKCNT; j++) {
			char *addr = chunk_addr[j]+start_ofs[j];
			int sz = CHUNKSZ-start_ofs[j]-end_ofs[j];
			sum2 = func(addr,addr+BUFSZ/2,sz,sum2,&err,&err);
			//sum2 = func(chunk_addr[j],chunk_addr[j]+BUFSZ/2,
			//	CHUNKSZ,sum2,&err,&err);
		}
		after = rdtsc();
		if(sum != sum2) die("bad sum");
		if(before>after) die("timer overflow");
		after -= before;
		if(min > after) min = after;
		if(max < after) max = after;
	}

	printf("took %5lli max,%5lli min cycles per kb. sum=0x%08x\n",
		max / (BUFSZ/2 / 1024),
		min / (BUFSZ/2 / 1024),
		sum
	);
	return sum;
}


unsigned fold_csum(unsigned sum) {
	sum = (sum & 0xFFFF) + ((sum>>16) & 0xFFFF);
	if(sum >= 0x10000) sum = sum-0x10000+1;
	return sum;
}

void test_csum()
{
	unsigned sum;
	puts("csum tests:");

#define	TEST_CSUM(a) fold_csum(test_one_csum(a,#a))
	sum =  TEST_CSUM(kernel_csum	);
	sum == TEST_CSUM(kernel_csum	) || die("bad sum");
	sum == TEST_CSUM(kernel_csum	) || die("bad sum");
	sum == TEST_CSUM(kernel2_csum	) || die("bad sum");
	sum == TEST_CSUM(kernelpii_csum	) || die("bad sum");
	sum == TEST_CSUM(kernelpiipf_csum) || die("bad sum");
	sum == TEST_CSUM(kpf_csum	) || die("bad sum");
	sum == TEST_CSUM(kpf_csum	) || die("bad sum");
	sum == TEST_CSUM(kpf_csum	) || die("bad sum");
	sum == TEST_CSUM(kpf_csum	) || die("bad sum");
//	sum == TEST_CSUM(pfm_csum	) || die("bad sum");
//	sum == TEST_CSUM(pfm2_csum	) || die("bad sum");
#undef TEST_CSUM
}

void test_copy()
{
	unsigned sum;
	puts("copy tests:");

#define	TEST_COPY(a) fold_csum(test_one_copy(a,#a))
	sum =  TEST_COPY(kernel_copy	);
	sum == TEST_COPY(kernel_copy	) || die("bad sum");
	sum == TEST_COPY(kernel_copy	) || die("bad sum");
	sum == TEST_COPY(kernelpii_copy	) || die("bad sum");
	sum == TEST_COPY(kernelpiipf_copy	) || die("bad sum");
	sum == TEST_COPY(kpf_copy	) || die("bad sum");
	sum == TEST_COPY(kpf_copy	) || die("bad sum");
	sum == TEST_COPY(kpf_copy	) || die("bad sum");
	sum == TEST_COPY(kpf_copy	) || die("bad sum");
//	sum == TEST_COPY(ntq_copy	) || die("bad sum");
//	sum == TEST_COPY(ntqpf_copy	) || die("bad sum");
//	sum == TEST_COPY(ntqpf2_copy	) || die("bad sum");
//	sum == TEST_COPY(ntqpfm_copy	) || die("bad sum");
#undef TEST_COPY
}

int main()
{
	char *buffer_raw;
	printf("Csum benchmark program. Buffer: %i Mb (%i). Chunks: %i kb (%i)\n"
		"Start/end offset in chunk: random from 0 to %i/%i\n"
		"Each test tried %i times, max and min CPU cycles are reported.\n"
		"Please disregard max values. They are due to system interference only.\n",
		BUFSZ/1024/1024, BUFSZ,
		CHUNKSZ/1024, CHUNKSZ,
		MAX_START_OFS,MAX_END_OFS,
		TRY_TIMES
	);

        buffer_raw = malloc(BUFSZ+BUF_ALIGN);
        if(!buffer_raw) die("Malloc failed");
        buffer = (char*) ((((int)buffer_raw)+(BUF_ALIGN-1)) & (~(BUF_ALIGN-1)));

	prepare();
	test_csum();
	test_copy();

	puts("Done");
	free(buffer_raw);
	return 0;
}
